//===========================================================================
// Released under Anonymity
// Feel free to embrace and extend this code!
// 11/16/02
// 
// ResourceExtractor.cpp
// 
// Implementation of an object that can extract resources 
// from a larger file
//===========================================================================

//===========================================================================
// Included Files
//===========================================================================
#include "projectpch.h"
#include "ResourceExtractor.h"

//===========================================================================
// Definition of Method: CResourceExtractor::GotoState(State* pNewState)
// 
// Transitions the state machine to a new state and calls beginState on the
// new state. Note no validation is done on the state transition. E.g. any 
// state can go to any state.  Future extensions of this implementation of
// the State pattern should probably include transition validation.
// (espcially useful if the machine gets large or nested). 
//===========================================================================
void CResourceExtractor::GotoState(State* pNewState)
{
	if (m_pCurrentState)
		delete m_pCurrentState;
	m_pCurrentState = pNewState;
	m_pCurrentState->beginState();
}	

//===========================================================================
// Definition of Method: CResourceExtractor::getOutFilename()
// 
// Generates a filename for a new resource file to be created based on an 
// incremented counter. It places that filename into m_szCurrFileName.
//
// Returns: a pointer to the character buffer holding the newly generated
// filename.
//===========================================================================
char* CResourceExtractor::getOutFilename()
{
	sprintf(m_szCurrFileName,"%s\\%lu.%s",m_szOutputDirectory,m_outfilenum++,m_szExtension);
	return m_szCurrFileName;
}

//===========================================================================
// Definition of Constructor: CResourceExtractor::CResourceExtractor
//                               (char* filename, char* extension,
//                                char* outputDirectory, char* startSentinel,
//                                int startSentinelLen,char* endSentinel,
//                                int endSentinelLen,bool dotrail,
//                                unsigned char trailbyte,
//                                IResourceExtractorListener* pListener)
// 
// Constructs a ResourceExtractor. Allocates space for several strings
// and connects to the listener (if the listener pointer is not-null)
//===========================================================================
CResourceExtractor::CResourceExtractor(char* filename, char* extension,  char* outputDirectory, char* startSentinel, int startSentinelLen,char* endSentinel, int endSentinelLen,bool dotrail,unsigned char trailbyte,IResourceExtractorListener* pListener)
:m_pfile(NULL)
,m_poutfile(NULL)
,m_numTrailingBytes(0)
,m_szStartCompareString(NULL)
,m_szEndCompareString(NULL)
,m_fp(0)
,m_fpend(0)
,m_pCurrentState(0)
,m_outfilenum(0)
,m_szOutputDirectory(NULL)
,m_szExtension(NULL)
,m_FilePos(0)
,m_CurrFileSize(0)
,m_FileSize(0)
,m_szMainFileName(NULL)
,m_bExtractorOK(true)
// the next few are initialized from constructor arguments
,m_output(pListener) 
,m_bDoTrailingBytes(dotrail)
,m_cTrailingByte(trailbyte)
,m_startcomparelen(startSentinelLen)
,m_endcomparelen(endSentinelLen)
{
	assert(filename);
	assert(extension);
	assert(outputDirectory);
	assert(startSentinel);
	assert(endSentinel);

		
	m_szStartCompareString=new char[startSentinelLen];
	assert(m_szStartCompareString);
	memcpy(m_szStartCompareString,startSentinel,startSentinelLen);

	m_szEndCompareString=new char[endSentinelLen];
	assert(m_szEndCompareString);
	memcpy(m_szEndCompareString,endSentinel,endSentinelLen);

	m_szExtension=new char[strlen(extension)+1];
	assert(m_szExtension);
	strcpy(m_szExtension,extension);

	m_szOutputDirectory=new char[strlen(outputDirectory)+1];
	assert(m_szOutputDirectory);
	strcpy(m_szOutputDirectory,outputDirectory);
	mkdir(m_szOutputDirectory);

	m_szMainFileName=new char[strlen(filename)+1];
	assert(m_szMainFileName);
	strcpy(m_szMainFileName,filename);
	
	memset(m_szErrorMessage,0,200); //should use #define for size... evil.

	m_pfile=fopen(m_szMainFileName,"rb");
	if (!m_pfile)
	{
		fatalExtractorError("Could not open main file");
	}
	else
	{
		fseek(m_pfile,0,SEEK_END);
		fgetpos(m_pfile,&m_fpend);
		fseek(m_pfile,0,SEEK_SET);
		assert(m_fpend>0);
		m_FileSize=(size_t)m_fpend-(size_t)m_fp;

		if (m_output)
			m_output->onMainFileOpen(m_szMainFileName,m_FileSize);
	}
}

//===========================================================================
// Definition of Method: CResourceExtractor::go()
// 
// The main scanning function of the ResourceExtractor.  Transitions to the
// state SearchingForStart and starts processing bytes until the the end of
// the file has been reached.
//===========================================================================
void CResourceExtractor::fatalExtractorError(const char* szError)
{
	//debugging code
	assert(m_bExtractorOK && "TWO FATAL EXTRACTOR ERRORS! BAD!");
	
	m_bExtractorOK=false;
	strcpy(m_szErrorMessage,szError);
	//disconnect our listener. 
	if (m_output)
		m_output->disconnect();

	//more debugging code
#ifdef _DEBUG
	printf("Error: %s\n",m_szErrorMessage);
#endif
	//assert(!"EXTRACTOR ENCOUNTERED FATAL ERROR: ");
}
//===========================================================================
// Definition of Method: CResourceExtractor::go()
// 
// The main scanning function of the ResourceExtractor.  Transitions to the
// state SearchingForStart and starts processing bytes until the the end of
// the file has been reached.
//===========================================================================
void CResourceExtractor::go()
{
	if (!isOK()) return;

	GotoState(new SearchingForStart(this));
	assert(m_pCurrentState);
	while (!feof(m_pfile))
	{
		if (!isOK()) return;

		fgetpos(m_pfile,&m_fp);
		int c=fgetc(m_pfile);
		processByte(c);
	}
}

//===========================================================================
// Definition of Method: CResourceExtractor::processByte(int c)
// 
// This base version of processByte first tells the listener that the
// ResourceExtractor is processing a byte, then defers to the current state
// to do the actual processing of the byte then increments the current file 
// position index.
//===========================================================================
void CResourceExtractor::processByte(int c)
{
	//technically should let each state determine if
	//it wants to notify listeners about each byte and increment the fileposition
	//but this is easier for now
	if (m_output)
		m_output->onByte(c,m_FilePos);
	m_pCurrentState->processByte(c);
	++m_FilePos;
}

//===========================================================================
// Definition of Method: CResourceExtractor::backTrack(int numChars)
// 
// Backtracks the given number of characters in the main file stream.
// slightly dangerous because it doesn't tell the current state.  and if you
// backtrack past any bytes "of interest" then the state machine may get
// all screwed up.
//===========================================================================
void CResourceExtractor::backTrack(int numChars)
{
	m_FilePos-=numChars;
	fseek(m_pfile,-numChars,SEEK_CUR);
}

//===========================================================================
// Definition of Destructor: CResourceExtractor::~CResourceExtractor()
// 
// Notify listener that the main file is closing (and basically that this
// extraction session has ended) and then deallocate all buffers used
// by this object
//===========================================================================
CResourceExtractor::~CResourceExtractor()
{
	if (m_pfile)
	{
		fclose(m_pfile);
		if (m_output)
			m_output->onMainFileClose(m_szMainFileName);
	}


	if(m_szMainFileName) delete [] m_szMainFileName;	
	if(m_szOutputDirectory) delete [] m_szOutputDirectory;
	if(m_szExtension) delete [] m_szExtension;
	if(m_szStartCompareString) delete [] m_szStartCompareString;
	if(m_szEndCompareString) delete [] m_szEndCompareString;	
}

