//===========================================================================
// Released under Anonymity
// Feel free to embrace and extend this code!
// 11/16/02
// 
// ResourceExtractorStatusOutput.cpp
// 
// Implementation of a status output object that models the 
// Reader/Writer problem. 
//===========================================================================

//===========================================================================
// Included Files
//===========================================================================
#include "projectpch.h"
#include "ResourceExtractorStatusOutput.h"

//===========================================================================
// Defines and Global Data
//===========================================================================
#define KBYTE 1024
#define MBYTE KBYTE*KBYTE
const int CResourceExtractorStatusOutput::c_printcountThreshold=50*KBYTE;
const int CResourceExtractorStatusOutput::c_printcountThreshold2=10*MBYTE;

//===========================================================================
// Definition of Procedure: CResourceExtractorStatusOutput::PrinterThread
//                                                      (void* pStatusOutput)
// 
// This method is the threadproc for the consumer thread of the characters
// put into the queue by the portion of the status output object that runs in
// the extractor's thread. (got that?)
//===========================================================================
void CResourceExtractorStatusOutput::PrinterThread(void* pStatusOutput)
{
	CResourceExtractorStatusOutput* pOutput=(CResourceExtractorStatusOutput*)pStatusOutput;
	while(1)
	{
		char sz[500];
		int i=0;
		//wait for at least one buffer element to be full
		WaitForSingleObject(pOutput->hDataAvailable,INFINITE);
		{
			//wait for and lock the queueAccess section in each loop
			//we care less about the performance of this thread, it is
			//just the output thread. so make sure to leave the critical 
			//section frequently to give the extractor thread as much 
			//processing time as possible.
			while(pOutput->m_queue.size() >0 && i < 500)
			{
				EnterCriticalSection(&pOutput->queueAccess);
					sz[i++]=pOutput->m_queue.front();
					pOutput->m_queue.pop();
				LeaveCriticalSection(&pOutput->queueAccess);
			}				
		}
		
		for (int j=0;j<i;++j)
		{
			if (sz[j]==(char)0xff)
			{
				//EOF means return from this thread
				return;
			}
			else
			{
				//print the character
				fputc(sz[j],stdout);
			}
		}
	}
	return;
}

//===========================================================================
// Definition of Constructor: CResourceExtractorStatusOutput::
//                                           CResourceExtractorStatusOutput()
//
// This constructor creates all the synchronization primitives that this
// object uses, and starts the consumer thread.
//===========================================================================
CResourceExtractorStatusOutput::CResourceExtractorStatusOutput()
:m_bExtracting(false)
,m_nextthreshold(c_printcountThreshold)
,m_nextthreshold2(c_printcountThreshold2)
{
	InitializeCriticalSection(&queueAccess);
	hDataAvailable=CreateSemaphore(NULL,0,500000,"Local\\ProjectXDataAvailable");
	m_hPrinterThread = (HANDLE) _beginthread(&PrinterThread,0,this);
}

//===========================================================================
// Definition of Destructor: CResourceExtractorStatusOutput::
//                                          ~CResourceExtractorStatusOutput()
//
// This destructor waits for the consumer thread to terminate then destroys
// the syncronization primitives used by this object
//===========================================================================
CResourceExtractorStatusOutput::~CResourceExtractorStatusOutput()
{
	//wait for the consumer thread to terminate
	WaitForSingleObject(m_hPrinterThread,10000);
	DeleteCriticalSection(&queueAccess);
	CloseHandle(hDataAvailable);
}

//===========================================================================
// Definition of Method: CResourceExtractorStatusOutput::onByte
//                                             (int byteValue,size_t byteNum)
//
// If this byte index is a multiple of the first threshold, then it produces
// a one character progress marker.  If the byte index is a multiple of the
// second threshold, then this method produces a status message advising
// the user of how much data has been processed and how much there is left
// to go.
//===========================================================================
void CResourceExtractorStatusOutput::onByte(int byteValue,size_t byteNum)
{
	if (byteNum >= m_nextthreshold)
	{
		printStatusMarker();
		m_nextthreshold+=c_printcountThreshold;
	}
	if (byteNum >= m_nextthreshold2)
	{
		printProgressMessage(byteNum);
		m_nextthreshold2+=c_printcountThreshold2;
	}
}

//===========================================================================
// Definition of Method: CResourceExtractorStatusOutput::onFileOpen
//                                         (char* szFilePath,size_t position)
//
// Produces a message advising the user that a resource has been found and
// a file is opening into which the resource will be extracted.
//===========================================================================
void CResourceExtractorStatusOutput::onFileOpen(char* szFilePath,size_t position)
{
	char sz[500];
	sprintf(sz,"\nFound file at offset %x!  extracting data into %s\n",position,szFilePath);
	produce(sz);
	m_bExtracting=true;
}

//===========================================================================
// Definition of Method: CResourceExtractorStatusOutput::onFileClose
//                                  (char* szFilePath, size_t position,
//                                  size_t numBytes, size_t numTrailingBytes)
//
// Produces a message advising the user that a resource has been completely
// extracted and the new file produced is now closing. Position is the 
// current position in the main file, numBytes is the size of the resource
// file, numTrailingBytes is the number of bytes at the end of the resource
// file that matched the trailing byte specifier.
//===========================================================================
void CResourceExtractorStatusOutput::onFileClose(char* szFilePath, size_t position,size_t numBytes, size_t numTrailingBytes)
{
	char sz[500];
	sprintf(sz,"\nClosing %s, at offset %x! (wrote %lu bytes, %lu trailing bytes)\n",szFilePath,position,numBytes,numTrailingBytes);
	produce(sz);
	m_bExtracting=false;
}

//===========================================================================
// Definition of Method: CResourceExtractorStatusOutput::onMainFileOpen
//                                   (char* szMainFileName,size_t totalBytes)
//
// Produces a message advising the user that the main file has opened and
// scanning has commenced.
//===========================================================================
void CResourceExtractorStatusOutput::onMainFileOpen(char* szMainFileName,size_t totalBytes)
{
	m_totBytes=totalBytes;
	produce("Opened ");
	produce(szMainFileName);
	produce(" for processing!\n");
}

//===========================================================================
// Definition of Method: CResourceExtractorStatusOutput::onMainFileClose
//                                                         (char* szFileName)
//
// Produces a message advising the user that the main file has closed and
// scanning has endedd.
//===========================================================================
void CResourceExtractorStatusOutput::onMainFileClose(char* szFileName)
{
	produce("Closed ");
	produce(szFileName);
	//signal for an end to the printer thread
	produce((char)0xff);
}

//===========================================================================
// Definition of Method: CResourceExtractorStatusOutput::printStatusMarker()
//
// Produces a one character progress marker.
//===========================================================================
void CResourceExtractorStatusOutput::printStatusMarker()
{
	if (m_bExtracting)
		produce('.');
	else
		produce('-');
}

//===========================================================================
// Definition of Method: CResourceExtractorStatusOutput::printProgressMessage
//                                                          (size_t numBytes)
//
// Produces a progress message advising the user how much of the main file
// has been scanned and how much is left
//===========================================================================
void CResourceExtractorStatusOutput::printProgressMessage(size_t numBytes)
{
	double percent = numBytes/m_totBytes;
	char sz[500];
	sprintf(sz,"\n%lu bytes (%5.2f%%) processed\n",numBytes,percent);
}

//===========================================================================
// Definition of Method: CResourceExtractorStatusOutput::produce(char* sz)
//
// Places a string of characters onto the character queue for consumption
// by the consumer/printer thread.
//===========================================================================
void CResourceExtractorStatusOutput::produce(char* sz)
{
	int i=0;
	//get queue access to batch add stuff
	while(*sz)
	{
		EnterCriticalSection(&queueAccess);
		m_queue.push(*sz++);
		LeaveCriticalSection(&queueAccess);
		++i;
	}
	//increase the count of full buffer elements
	ReleaseSemaphore(hDataAvailable,i,NULL);
}

//===========================================================================
// Definition of Method: CResourceExtractorStatusOutput::produce(char c)
//
// Places a character onto the character queue for consumption by the 
// consumer/printer thread.
//===========================================================================
void CResourceExtractorStatusOutput::produce(char c)
{
	EnterCriticalSection(&queueAccess);
	m_queue.push(c);
	LeaveCriticalSection(&queueAccess);
	//increase the count of full buffer elements
	ReleaseSemaphore(hDataAvailable,1,NULL);
}

//===========================================================================
// Definition of Method: CResourceExtractorStatusOutput::produce(char c)
//
// Called when the listener is disconnected prematurely.  If not called then
// this listener will shutdown the printer thread when the end of the main 
// file is reached.
//===========================================================================
void CResourceExtractorStatusOutput::disconnect()
{
	produce((char)0xff);
}
